@@ -60,7 +60,7 @@ static GObjectClass *parent_class;
   		g_object_unref (object);									   \
   		goto beach;													   \
   	}																   \
-	
+
 #define CONVERT_LIST(py_files, files)                                  \
 	{                                                                  \
 		GList *l;                                                      \
@@ -107,7 +107,26 @@ static GObjectClass *parent_class;
             Py_DECREF(py_item);                                        \
     	}                                                              \
     }
-    
+
+
+static void
+free_pygobject_data(gpointer data, gpointer user_data)
+{
+	/* Some NautilusFile objects are cached and not freed until nautilus
+		itself is closed.  Since PyGObject stores data that must be freed by
+		the Python interpreter, we must always free it before the interpreter
+		is finalized. */
+	g_object_set_data((GObject *)data, "PyGObject::instance-data", NULL);
+}
+
+static void
+free_pygobject_data_list(GList *list)
+{
+	if (list == NULL)
+		return;
+		
+	g_list_foreach(list, (GFunc)free_pygobject_data, NULL);
+}
 
 #define METHOD_NAME "get_property_pages"
 static GList *
@@ -133,7 +152,6 @@ nautilus_python_object_get_property_pages (NautilusPropertyPageProvider *provide
 	HANDLE_LIST(py_ret, NautilusPropertyPage, "nautilus.PropertyPage");
 	
  beach:
- 	Py_XDECREF(py_files);
 	Py_XDECREF(py_ret);
 	pyg_gil_state_release(state);
     return ret;
@@ -182,8 +200,6 @@ nautilus_python_object_get_widget (NautilusLocationWidgetProvider *provider,
 	ret = (GtkWidget *)g_object_ref(py_ret_gobj->obj);
 
  beach:
- 	Py_XDECREF(py_uri);
- 	Py_XDECREF(py_ret_gobj);
 	Py_XDECREF(py_ret);
 	pyg_gil_state_release(state);
 	return ret;
@@ -206,22 +222,40 @@ nautilus_python_object_get_file_items (NautilusMenuProvider *provider,
     GList *ret = NULL;
     PyObject *py_ret = NULL, *py_files;
 	PyGILState_STATE state = pyg_gil_state_ensure();
+	PyObject *provider_version = NULL;
 	
   	debug_enter();
 
-	CHECK_OBJECT(object);
-	CHECK_METHOD_NAME(object->instance);
+	CHECK_OBJECT(object);	
 
-	CONVERT_LIST(py_files, files);
+	if (PyObject_HasAttrString(object->instance, "get_file_items_full"))
+	{
+		CONVERT_LIST(py_files, files);
+		py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX "get_file_items_full",
+									 "(NNN)",
+									 pygobject_new((GObject *)provider), 
+									 pygobject_new((GObject *)window), 
+									 py_files);
+	}
+	else if (PyObject_HasAttrString(object->instance, "get_file_items"))
+	{
+		CONVERT_LIST(py_files, files);
+		py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX METHOD_NAME,
+									 "(NN)", 
+									 pygobject_new((GObject *)window), 
+									 py_files);
+	}
+	else
+	{
+		goto beach;
+	}
 
-    py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX METHOD_NAME,
-								 "(NN)", pygobject_new((GObject *)window), py_files);
 	HANDLE_RETVAL(py_ret);
 
 	HANDLE_LIST(py_ret, NautilusMenuItem, "nautilus.MenuItem");
 
  beach:
- 	Py_XDECREF(py_files);
+ 	free_pygobject_data_list(files);
 	Py_XDECREF(py_ret);
 	pyg_gil_state_release(state);
     return ret;
@@ -237,23 +271,38 @@ nautilus_python_object_get_background_items (NautilusMenuProvider *provider,
 	NautilusPythonObject *object = (NautilusPythonObject*)provider;
     GList *ret = NULL;
     PyObject *py_ret = NULL;
-	PyGILState_STATE state = pyg_gil_state_ensure();                                    \
+	PyGILState_STATE state = pyg_gil_state_ensure();
 	
   	debug_enter();
 
 	CHECK_OBJECT(object);
-	CHECK_METHOD_NAME(object->instance);
 
-    py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX METHOD_NAME,
-								 "(NN)",
-								 pygobject_new((GObject *)window),
-								 pygobject_new((GObject *)file));
-								 
+	if (PyObject_HasAttrString(object->instance, "get_background_items_full"))
+	{
+		py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX "get_background_items_full",
+									 "(NNN)",
+									 pygobject_new((GObject *)provider),
+									 pygobject_new((GObject *)window),
+									 pygobject_new((GObject *)file));
+	}
+	else if (PyObject_HasAttrString(object->instance, "get_background_items"))
+	{
+		py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX METHOD_NAME,
+									 "(NN)",
+									 pygobject_new((GObject *)window),
+									 pygobject_new((GObject *)file));
+	}
+	else
+	{
+		goto beach;
+	}
+
 	HANDLE_RETVAL(py_ret);
 
 	HANDLE_LIST(py_ret, NautilusMenuItem, "nautilus.MenuItem");
 	
  beach:
+	free_pygobject_data(file, NULL);
 	Py_XDECREF(py_ret);
 	pyg_gil_state_release(state);
     return ret;
@@ -269,22 +318,38 @@ nautilus_python_object_get_toolbar_items (NautilusMenuProvider *provider,
 	NautilusPythonObject *object = (NautilusPythonObject*)provider;
     GList *ret = NULL;
     PyObject *py_ret = NULL;
-	PyGILState_STATE state = pyg_gil_state_ensure();                                    \
+	PyGILState_STATE state = pyg_gil_state_ensure();
 	
   	debug_enter();
   	
 	CHECK_OBJECT(object);
-	CHECK_METHOD_NAME(object->instance);
 
-    py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX METHOD_NAME,
-								 "(NN)",
-								 pygobject_new((GObject *)window),
-								 pygobject_new((GObject *)file));
+	if (PyObject_HasAttrString(object->instance, "get_toolbar_items_full"))
+	{
+		py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX "get_toolbar_items_full",
+									 "(NNN)",
+									 pygobject_new((GObject *)provider),
+									 pygobject_new((GObject *)window),
+									 pygobject_new((GObject *)file));
+	}
+	else if (PyObject_HasAttrString(object->instance, "get_toolbar_items"))
+	{
+		py_ret = PyObject_CallMethod(object->instance, METHOD_PREFIX METHOD_NAME,
+									 "(NN)",
+									 pygobject_new((GObject *)window),
+									 pygobject_new((GObject *)file));
+	}
+	else
+	{
+		goto beach;
+	}
+	
 	HANDLE_RETVAL(py_ret);
 
 	HANDLE_LIST(py_ret, NautilusMenuItem, "nautilus.MenuItem");
 	
  beach:
+ 	free_pygobject_data(file, NULL);
 	Py_XDECREF(py_ret);
 	pyg_gil_state_release(state);
     return ret;
@@ -339,7 +404,21 @@ static void
 nautilus_python_object_cancel_update (NautilusInfoProvider 		*provider,
 									  NautilusOperationHandle 	*handle)
 {
+	NautilusPythonObject *object = (NautilusPythonObject*)provider;
+	PyGILState_STATE state = pyg_gil_state_ensure();
+
   	debug_enter();
+
+	CHECK_OBJECT(object);
+	CHECK_METHOD_NAME(object->instance);
+
+    PyObject_CallMethod(object->instance,
+								 METHOD_PREFIX METHOD_NAME, "(NN)",
+								 pygobject_new((GObject*)provider),
+								 pyg_pointer_new(G_TYPE_POINTER, handle));
+
+ beach:
+	pyg_gil_state_release(state);
 }
 #undef METHOD_NAME
 
@@ -354,18 +433,33 @@ nautilus_python_object_update_file_info (NautilusInfoProvider 		*provider,
     NautilusOperationResult ret = NAUTILUS_OPERATION_COMPLETE;
     PyObject *py_ret = NULL;
 	PyGILState_STATE state = pyg_gil_state_ensure();                                    \
-	
+
   	debug_enter();
 
 	CHECK_OBJECT(object);
-	CHECK_METHOD_NAME(object->instance);
 
-    py_ret = PyObject_CallMethod(object->instance,
-								 METHOD_PREFIX METHOD_NAME, "(N)",
-								 pygobject_new((GObject*)file));
+	if (PyObject_HasAttrString(object->instance, "update_file_info_full"))
+	{
+		py_ret = PyObject_CallMethod(object->instance,
+									 METHOD_PREFIX "update_file_info_full", "(NNNN)",
+									 pygobject_new((GObject*)provider),
+									 pyg_pointer_new(G_TYPE_POINTER, *handle),
+									 pyg_boxed_new(G_TYPE_CLOSURE, update_complete, TRUE, TRUE),
+									 pygobject_new((GObject*)file));
+	}
+	else if (PyObject_HasAttrString(object->instance, "update_file_info"))
+	{
+		py_ret = PyObject_CallMethod(object->instance,
+									 METHOD_PREFIX METHOD_NAME, "(N)",
+									 pygobject_new((GObject*)file));
+	}
+	else
+	{
+		goto beach;
+	}
+	
 	HANDLE_RETVAL(py_ret);
 
-
 	if (!PyInt_Check(py_ret))
 	{
 		PyErr_SetString(PyExc_TypeError,
@@ -376,6 +470,7 @@ nautilus_python_object_update_file_info (NautilusInfoProvider 		*provider,
 	ret = PyInt_AsLong(py_ret);
 	
  beach:
+ 	free_pygobject_data(file, NULL);
 	Py_XDECREF(py_ret);
 	pyg_gil_state_release(state);
     return ret;
